from enum import Enum
from abc import ABC, abstractmethod


class TableStrategyTypes(Enum):
    TRUNCATE = "TRUNCATE"
    UPDATE_COLUMNS = "UPDATE_COLUMNS"
    DELETE = "DELETE"

    @staticmethod
    def from_value(string):
        try:
            return TableStrategyTypes(string.upper())
        except ValueError:
            return None


class TableStrategy(ABC):
    def __init__(self, table_name, schema=None):
        self.table_name = table_name
        self.schema = schema

    @property
    def qualified_name(self):
        return f"{self.schema}.{self.table_name}" if self.schema else self.table_name

class DeleteTableStrategy(TableStrategy):
    strategy_type = TableStrategyTypes.DELETE

class TruncateTableStrategy(TableStrategy):
    strategy_type = TableStrategyTypes.TRUNCATE


class UpdateColumnsTableStrategy(TableStrategy):
    strategy_type = TableStrategyTypes.UPDATE_COLUMNS

    def __init__(self, table_name, column_strategies, schema=None):
        super().__init__(table_name=table_name, schema=schema)
        self.__column_strategies = []
        for column_strategy in column_strategies:
            self.__column_strategies.append(column_strategy)

    def group_by_where(self):
        """
        returns a map of columns, grouped in a map of where conditions
        :return:
        """
        grouped_columns = {}

        for column_strategy in self.__column_strategies:
            where_condition = column_strategy.where_condition
            if where_condition not in grouped_columns:
                grouped_columns[where_condition] = {}

            grouped_columns[where_condition][column_strategy.column_name] = column_strategy

        return grouped_columns

    @property
    def column_strategies(self):
        return self.__column_strategies

    def get_column_strategies(self):
        """Legacy, replace with property"""
        return self.column_strategies



import os
from yaml import safe_load as load_yaml
from json import loads as load_json


class UnknownConfigTypeError(Exception):
    pass


def read_config(file_path):
    name, ext = os.path.splitext(os.path.basename(file_path))

    with open(file_path, "r") as config_data:
        if ext == ".yml" or ext == ".yaml":
            return load_yaml(config_data)
        elif ext == ".json":
            return load_json(config_data.read())
        else:
            raise UnknownConfigTypeError(f"Unknown config type {ext} from {file_path}")


from pynonymizer.database.provider import DatabaseProvider
from pynonymizer.database.provider import SEED_TABLE_NAME
from pynonymizer.strategy.update_column import UpdateColumnStrategyTypes
from pynonymizer.strategy.table import TableStrategyTypes
from pynonymizer.database.exceptions import UnsupportedColumnStrategyError, UnsupportedTableStrategyError, DependencyError
from pynonymizer.fake import FakeDataType

import math
import logging
from tqdm import tqdm
from pathlib import PureWindowsPath, PurePosixPath
import re


_FAKE_COLUMN_TYPES = {
    FakeDataType.STRING: "VARCHAR(MAX)",
    FakeDataType.DATE: "DATE",
    FakeDataType.DATETIME: "DATETIME",
    FakeDataType.INT: "INT"
}

_LOCAL_SERVER = "127.0.0.1"
_DEFAULT_PORT = "1433"

def _extract_driver_version(driver):
    try:
        return int( re.findall(r"\d+", driver)[0] )
    except IndexError:
        return 0

class MsSqlProvider(DatabaseProvider):
    """
    A pyodbc-based MSSQL provider.
    """

    logger = logging.getLogger(__name__)

    # stats value for restore/backup command: Report progress every X percent
    # A lower value means MORE resultssets / more frequent updates from the backup command.
    # Values lower than 5 often yield unreliable results on smaller backups
    __STATS = 5

    def __init__(self, db_host, db_user, db_pass, db_name, db_port=None, seed_rows=None, backup_compression=False, driver=None):
        # import here for fast-failiness
        import pyodbc

        db_host = db_host or _LOCAL_SERVER
        db_port = db_port or _DEFAULT_PORT
        driver = driver or self.__detect_driver()

        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port

        if seed_rows is None:
            seed_rows = 150

        self.seed_rows = int(seed_rows)

        self.__conn = None
        self.__db_conn = None
        self.__backup_compression = backup_compression
        self.__driver = driver

    def __detect_driver(self):
        import pyodbc
        
        ms_drivers = [i for i in pyodbc.drivers() if "sql server" in i.lower()]
        if len(ms_drivers) < 1:
            raise DependencyError("odbc", "Failed to detect any ODBC drivers on this system.")

        if len(ms_drivers) > 1:
            self.logger.debug("multiple drivers detected for mssql: %s", ms_drivers)

        # Sort by the highest number (like, ODBC driver 14 for SQL server)
        return sorted(ms_drivers, key=_extract_driver_version, reverse=True)[0]

    def __require_local_server(self):
        if self.db_host != _LOCAL_SERVER:
            raise DependencyError("db_host", "This operation does not support remote servers due to backup file "
                                                "location requirements. You must omit db_host from your configuration "
                                                "and run pynonymizer on the same server as the database.")

    def __connection(self):
        import pyodbc
        """a lazy-evaluated connection"""
        if self.__conn is None:
            self.__conn = pyodbc.connect(
                driver=f"{{{self.__driver}}}",
                server=f"{self.db_host},{self.db_port}",
                uid=self.db_user,
                pwd=self.db_pass,
                autocommit=True
            )

        return self.__conn

    def __db_connection(self):
        import pyodbc
        """a lazy-evaluated db-specific connection"""
        if self.__db_conn is None:
            self.__db_conn = pyodbc.connect(
                driver=f"{{{self.__driver}}}",
                database=self.db_name,
                server=f"{self.db_host},{self.db_port}",
                uid=self.db_user,
                pwd=self.db_pass,
                autocommit=True
            )

        return self.__db_conn

    def __execute(self, *args, **kwargs):
        return self.__connection().execute(*args, **kwargs)

    def __db_execute(self, *args, **kwargs):
        return self.__db_connection().execute(*args, **kwargs)

    def __get_path(self, filepath):
        if "\\" in filepath:
            return PureWindowsPath(filepath)
        else:
            return PurePosixPath(filepath)

    def __get_default_datafolder(self):
        """
        Locate the default data folder using the `model` database location
        It's possible that the model database is not the currently set default, i.e if it's been changed after install
        The solution to this would be to make a new database and then perform the below check on that instead.
        See https://blogs.technet.microsoft.com/sqlman/2009/07/19/tsql-script-determining-default-database-file-log-path/

        However, this seems like a heavyweight solution for what is essentially a tsql-writeable tempfolder, so
        checking the model db seems like a good 'boring' solution
        :return: Default data directory e.g. "C:\\DATA"
        """
        datafile = self.__execute("""
        SELECT physical_name
        FROM sys.master_files mf
        INNER JOIN sys.[databases] d
        ON mf.[database_id] = d.[database_id]
        WHERE d.[name] = 'model' AND type = 0
        """).fetchone()[0]

        return self.__get_path(datafile).parent

    def __get_default_logfolder(self):
        """
        Locate the default log folder using the `model` database location
        __get_default_datafolder: see for more info
        :return:
        """
        logfile = self.__execute("""
        SELECT physical_name
        FROM sys.master_files mf
        INNER JOIN sys.[databases] d
        ON mf.[database_id] = d.[database_id]
        WHERE d.[name] = 'model' AND type = 1
        """).fetchone()[0]

        return self.__get_path(logfile).parent

    def __get_file_moves(self, input_path):
        """
        Using RESTORE FILELISTONLY, get all the files in the backup that need to be moved to the local system for restore
        :return: a dict of file name: new destination
        """
        datadir = self.__get_default_datafolder()
        logdir = self.__get_default_logfolder()

        filelist = self.__execute(f"RESTORE FILELISTONLY FROM DISK = ?;", input_path).fetchall()

        move_file_map = {}
        for file in filelist:
            name = file[0]
            type = file[2].upper()
            filepath = self.__get_path(file[1])

            # log files can go into the default log directory, everything else can go into the data directory
            if type == "L":
                target_path = str(logdir.joinpath(f"{self.db_name}_{filepath.name}"))
            else:
                target_path = str(datadir.joinpath(f"{self.db_name}_{filepath.name}"))

            move_file_map[name] = target_path

        return move_file_map

    def __async_operation_progress(self, desc, cursor):
        # With STATS=x, we should recieve 100/x resultsets, provided the backup is slow enough.
        # With some databases, it will jump from y% to 100, so we'll only get <x nextset calls.
        # Even SSMS doesn't get informed (read: it's not my fault, blame microsoft)
        with tqdm(desc=desc, total=math.floor(100 / self.__STATS)) as progressbar:
            while cursor.nextset():
                progressbar.update()

            # finish the progress - less confusing than a dangling 40% progressbar
            progressbar.update(progressbar.total - progressbar.n)

    def __run_scripts(self, script_list, title=""):
        import pyodbc
        for i, script in enumerate(script_list):
            self.logger.info(f"Running {title} script #{i} \"{script[:50]}\"")
            cursor = self.__db_execute(script)
            results = None
            try:
                results = cursor.fetchall()
            except pyodbc.Error:
                pass
            self.logger.info(results)

    def __create_seed_table(self, qualifier_map):
        seed_column_lines = ["[{}] {}".format(name, _FAKE_COLUMN_TYPES[col.data_type]) for name, col in qualifier_map.items()]
        create_statement = "CREATE TABLE [{}]({});".format(SEED_TABLE_NAME, ",".join(seed_column_lines))

        self.__db_execute(create_statement)

    def __drop_seed_table(self):
        self.__db_execute("DROP TABLE IF EXISTS [{}];".format(SEED_TABLE_NAME))

    def __insert_seed_row(self, qualifier_map):
        column_list = ",".join(["[{}]".format(qualifier) for qualifier in qualifier_map])
        substitution_list = ",".join([" ?".format(qualifier) for qualifier in qualifier_map])
        value_list = [column.value for qualifier, column in qualifier_map.items()]

        statement = "INSERT INTO [{}]({}) VALUES ({});".format(SEED_TABLE_NAME, column_list, substitution_list)
        self.__db_execute(statement, value_list)

    def __seed(self, qualifier_map):
        for i in tqdm(range(0, self.seed_rows), desc="Inserting seed data", unit="rows"):
            self.__insert_seed_row(qualifier_map)

    def __get_column_subquery(self, column_strategy, table_name, column_name):
        if column_strategy.strategy_type == UpdateColumnStrategyTypes.EMPTY:
            return "('')"
        elif column_strategy.strategy_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL:
            return f"( SELECT CONCAT(NEWID(), '@', NEWID(), '.com') )"
        elif column_strategy.strategy_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN:
            return f"( SELECT NEWID() )"
        elif column_strategy.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE:
            column = f"[{column_strategy.qualifier}]"
            if column_strategy.sql_type:
                column = f"CAST({column} AS {column_strategy.sql_type})"
            # Add WHERE LIKE % OR NULL to make subquery correlated with outer table, therefore uncachable
            return f"( SELECT TOP 1 {column} FROM [{SEED_TABLE_NAME}] WHERE [{table_name}].[{column_name}] LIKE '%' OR [{table_name}].[{column_name}] IS NULL ORDER BY NEWID())"
        elif column_strategy.strategy_type == UpdateColumnStrategyTypes.LITERAL:
            return column_strategy.value
        else:
            raise UnsupportedColumnStrategyError(column_strategy)

    def create_database(self):
        self.logger.warning("MSSQL: create_database ignored, database will be created when restore_db is run")

    def drop_database(self):
        # force connection close so we can always drop the db: sometimes timing makes a normal drop impossible.
        self.__execute(f"ALTER DATABASE [{self.db_name}] SET SINGLE_USER WITH ROLLBACK IMMEDIATE;")
        self.__execute(f"DROP DATABASE IF EXISTS [{self.db_name}];")

    def anonymize_database(self, database_strategy):
        qualifier_map = database_strategy.fake_update_qualifier_map

        if len(qualifier_map) > 0:
            self.logger.info("creating seed table with %d columns", len(qualifier_map))
            self.__create_seed_table(qualifier_map)

            self.logger.info("Inserting seed data")
            self.__seed(qualifier_map)

        self.__run_scripts(database_strategy.before_scripts, "before")

        table_strategies = database_strategy.table_strategies
        self.logger.info("Anonymizing %d tables", len(table_strategies))

        with tqdm(desc="Anonymizing database", total=len(table_strategies)) as progressbar:
            for table_strategy in table_strategies:
                table_name = table_strategy.table_name
                schema_prefix = f"[{table_strategy.schema}]." if table_strategy.schema else ""

                if table_strategy.strategy_type == TableStrategyTypes.TRUNCATE:
                    progressbar.set_description("Truncating {}".format(table_name))
                    self.__db_execute("TRUNCATE TABLE {}[{}];".format(schema_prefix, table_name))

                elif table_strategy.strategy_type == TableStrategyTypes.DELETE:
                    progressbar.set_description("Deleting {}".format(table_name))
                    self.__db_execute("DELETE FROM {}[{}];".format(schema_prefix, table_name))

                elif table_strategy.strategy_type == TableStrategyTypes.UPDATE_COLUMNS:
                    progressbar.set_description("Anonymizing {}".format(table_name))
                    where_grouping = table_strategy.group_by_where()
                    total_wheres = len(where_grouping)

                    for i, (where, column_map) in enumerate(where_grouping.items()):
                        column_assignments = ",".join(["[{}] = {}".format(name, self.__get_column_subquery(column, table_name, name)) for name, column in column_map.items()])
                        where_clause = f" WHERE {where}" if where else ""
                        progressbar.set_description("Anonymizing {}: w[{}/{}]".format(table_name, i+1, total_wheres))
                        # Disable ANSI_WARNINGS to allow oversized fake data to be truncated without error
                        self.__db_execute("SET ANSI_WARNINGS off; UPDATE {}[{}] SET {}{}; SET ANSI_WARNINGS on;".format(schema_prefix, table_name, column_assignments, where_clause))

                else:
                    raise UnsupportedTableStrategyError(table_strategy)

                progressbar.update()

        self.__run_scripts(database_strategy.after_scripts, "after")

        self.logger.info("Dropping seed table")
        self.__drop_seed_table()

    def restore_database(self, input_path):
        self.__require_local_server()

        move_files = self.__get_file_moves(input_path)

        self.logger.info("Found %d files in %s", len(move_files), input_path )
        self.logger.debug(move_files)

        # get move statements and flatten pairs out so we can do the 2-param substitution
        move_clauses = ", ".join( ["MOVE ? TO ?"] * len(move_files) )
        move_clause_params = [item for pair in move_files.items() for item in pair]

        restore_cursor = self.__execute(f"RESTORE DATABASE ? FROM DISK = ? WITH {move_clauses}, STATS = ?;",
                                        [self.db_name, input_path, *move_clause_params, self.__STATS])

        self.__async_operation_progress("Restoring Database", restore_cursor)

    def dump_database(self, output_path):
        self.__require_local_server()

        with_options = []
        if self.__backup_compression:
            with_options.append("COMPRESSION")

        with_options_str = ",".join(with_options) + ", " if len(with_options) > 0 else ""

        dump_cursor = self.__execute(f"BACKUP DATABASE ? TO DISK = ? WITH {with_options_str}STATS = ?;", [self.db_name, output_path, self.__STATS])
        self.__async_operation_progress("Dumping Database", dump_cursor)



class PynonymizerError(Exception):
    """
    A base error class for this application.
    """
    pass


class ArgumentValidationError(PynonymizerError):
    def __init__(self, validation_messages):
        self.validation_messages = validation_messages


class DatabaseConnectionError(PynonymizerError):
    pass


from pynonymizer.strategy.table import TableStrategyTypes
from pynonymizer.strategy.update_column import UpdateColumnStrategyTypes


class DatabaseStrategy:
    def __init__(self, table_strategies=None, before_scripts=None, after_scripts=None):

        self.table_strategies = []
        self.before_scripts = []
        self.after_scripts = []

        for table_strat in table_strategies:
            self.table_strategies.append(table_strat)

        if before_scripts:
            for script in before_scripts:
                self.before_scripts.append(script)

        if after_scripts:
            for script in after_scripts:
                self.after_scripts.append(script)


    @property
    def scripts(self):
        """
        Deprecated - use before/after vars
        :return:
        """
        return {
            "before": self.before_scripts,
            "after": self.after_scripts
        }

    @property
    def fake_update_qualifier_map(self):
        column_strategies = {}
        for table_strategy in self.table_strategies:
            if table_strategy.strategy_type == TableStrategyTypes.UPDATE_COLUMNS:
                for column_strategy in table_strategy.column_strategies:
                    if column_strategy.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE:
                        column_strategies[column_strategy.qualifier] = column_strategy

        return column_strategies

    @property
    def all_column_strategies(self):
        column_strategies = {}
        for table_strategy in self.table_strategies:
            if table_strategy.strategy_type == TableStrategyTypes.UPDATE_COLUMNS:
                column_strategies += table_strategy.column_strategies

        return column_strategies

    def get_all_column_strategies(self):
        """LEGACY: move to property"""
        return self.all_column_strategies



from enum import Enum
from pynonymizer.fake import  UnsupportedFakeTypeError
from abc import ABC, abstractmethod
import hashlib

class UpdateColumnStrategyTypes(Enum):
    EMPTY = "EMPTY"
    UNIQUE_LOGIN = "UNIQUE_LOGIN"
    UNIQUE_EMAIL = "UNIQUE_EMAIL"
    FAKE_UPDATE = "FAKE_UPDATE"
    LITERAL = "LITERAL"

    @staticmethod
    def from_value(string):
        """
        a method to get an enum value from it's case-insensitive value.
        """
        try:
            return UpdateColumnStrategyTypes(string.upper())
        except ValueError:
            return None


# boilerplate abstract class for future use
class UpdateColumnStrategy(ABC):
    def __init__(self, column_name, where=None):
        self.column_name = column_name
        self.where_condition = where


class EmptyUpdateColumnStrategy(UpdateColumnStrategy):
    strategy_type = UpdateColumnStrategyTypes.EMPTY


class UniqueLoginUpdateColumnStrategy(UpdateColumnStrategy):
    strategy_type = UpdateColumnStrategyTypes.UNIQUE_LOGIN


class UniqueEmailUpdateColumnStrategy(UpdateColumnStrategy):
    strategy_type = UpdateColumnStrategyTypes.UNIQUE_EMAIL


class LiteralUpdateColumnStrategy(UpdateColumnStrategy):
    strategy_type = UpdateColumnStrategyTypes.LITERAL

    def __init__(self, column_name,  value, where=None):
        super().__init__(column_name, where)
        self.value = value


class FakeUpdateColumnStrategy(UpdateColumnStrategy):
    strategy_type = UpdateColumnStrategyTypes.FAKE_UPDATE

    def __init__(self, column_name, fake_column_generator, fake_type, where=None, fake_args=None, sql_type=None):
        fake_args = {} if fake_args is None else fake_args
        super().__init__(column_name, where)
        self.fake_type = fake_type
        self.fake_args = fake_args
        self.__fake_column_generator = fake_column_generator
        self.sql_type = sql_type

        if not fake_column_generator.supports(fake_type, fake_args):
            raise UnsupportedFakeTypeError(fake_type, fake_args)

    @property
    def qualifier(self):
        """
        Generate a deterministic qualifier for this fake update column strategy, so that it and it's args can be
        identified in cases like pre-generation of values
        e.g. file_path_depth_1
        :return:
        """
        sorted_args = "_".join( [f"{arg}_{value}" for arg, value in sorted(self.fake_args.items(), key=lambda item: item[0])] )
        args_hash = hashlib.md5(bytes(sorted_args, "utf8")).hexdigest()

        # keep the whole thing below 64chars for maximum database compatibility
        return (self.fake_type + (("_" + args_hash) if sorted_args else ""))[:64]


    @property
    def value(self):
        """
        Generate a value from the faker data this column represents
        :return:
        """
        return self.__fake_column_generator.get_value(self.fake_type, self.fake_args)

    @property
    def data_type(self):
        """
        get this generator's data type
        e.g. FakeDataTypes.STRING
        :return:
        """
        return self.__fake_column_generator.get_data_type(self.fake_type)



from datetime import date, datetime
from pynonymizer.database.exceptions import UnsupportedColumnStrategyError
from pynonymizer.strategy.update_column import UpdateColumnStrategyTypes
from pynonymizer.fake import FakeDataType
"""
All Static query generation functions
"""
_FAKE_COLUMN_TYPES = {
    FakeDataType.STRING: "TEXT",
    FakeDataType.DATE: "DATE",
    FakeDataType.DATETIME: "DATETIME",
    FakeDataType.INT: "INT"
}

# For preservation of unique values across versions of mysql, and this bug:
# https://bugs.mysql.com/bug.php?id=89474, use md5 based rand subqueries for unique values (rather than UUIDs)
_RAND_MD5 = "MD5(FLOOR((NOW() + RAND()) * (RAND() * RAND() / RAND()) + RAND()))"


def _get_sql_type(data_type):
    return _FAKE_COLUMN_TYPES[data_type]


def _get_column_subquery(seed_table_name, column_strategy):
    if column_strategy.strategy_type == UpdateColumnStrategyTypes.EMPTY:
        return "('')"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL:
        return f"( SELECT CONCAT({_RAND_MD5}, '@', {_RAND_MD5}, '.com') )"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN:
        return f"( SELECT {_RAND_MD5} )"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE:
        column = f"`{column_strategy.qualifier}`"
        if column_strategy.sql_type:
            column = f"CAST({column} AS {column_strategy.sql_type})"
        return f"( SELECT {column} FROM `{seed_table_name}` ORDER BY RAND() LIMIT 1)"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.LITERAL:
        return column_strategy.value
    else:
        raise UnsupportedColumnStrategyError(column_strategy)


def _escape_sql_value(value):
    """
    return a sql-ified version of a seed column's value
    Normally this defines the stringification of datatypes and escaping for strings
    """
    if isinstance(value, (str, datetime, date)):
        return "'" + str(value).replace("'", "''") + "'"
    else:
        return str(value)


def get_truncate_table(table_name):
    return f"SET FOREIGN_KEY_CHECKS=0; TRUNCATE TABLE `{table_name}`; SET FOREIGN_KEY_CHECKS=1;"

def get_delete_table(table_name):
    return f"DELETE FROM `{table_name}`;"


def get_create_seed_table(table_name, qualifier_map):
    if len(qualifier_map) < 1:
        raise ValueError("Cannot create a seed table with no columns")

    create_columns = [f"`{qualifier}` {_get_sql_type(strategy.data_type)}" for qualifier, strategy in qualifier_map.items()]

    return "CREATE TABLE `{}` ({});".format(table_name, ",".join(create_columns) )


def get_drop_seed_table(table_name):
    return f"DROP TABLE IF EXISTS `{table_name}`;"


def get_insert_seed_row(table_name, qualifier_map):

    column_names = ",".join( [f"`{qualifier}`" for qualifier in qualifier_map.keys()] )
    column_values = ",".join( [f"{_escape_sql_value(strategy.value)}" for strategy in qualifier_map.values()] )

    return "INSERT INTO `{}`({}) VALUES ({});".format(table_name, column_names, column_values)


def get_create_database(database_name):
    return f"CREATE DATABASE `{database_name}`;"


def get_drop_database(database_name):
    return f"DROP DATABASE IF EXISTS `{database_name}`;"


# TODO: this where-grouping behaviour should probably return to the provider, rather than implementing as q-gen logic
def get_update_table(seed_table_name, update_table_strategy):
    # group on where_condition
    # build lists of update statements based on the where
    output_statements = []
    where_update_statements = {}
    for where, column_map in update_table_strategy.group_by_where().items():
        where_update_statements[where] = []
        for column_name, column_strategy in column_map.items():
            where_update_statements[where].append("`{}` = {}".format(
                column_name,
                _get_column_subquery(seed_table_name, column_strategy))
            )

        assignments = ",".join( where_update_statements[where] )
        where_clause = f" WHERE {where}" if where else ""

        output_statements.append( f"UPDATE `{update_table_strategy.table_name}` SET {assignments}{where_clause};")

    return output_statements


def get_dumpsize_estimate(database_name):
    return "SELECT data_bytes " \
           "FROM (SELECT SUM(data_length) AS data_bytes FROM information_schema.tables WHERE table_schema = '{}') " \
           "AS data;".format(database_name)



import argparse
import dotenv
import os
import sys
import logging
from pynonymizer.pynonymize import ArgumentValidationError, DatabaseConnectionError, pynonymize, ProcessSteps
from pynonymizer import __version__


def create_parser():
    parser = argparse.ArgumentParser(prog="pynonymizer", description="A tool for writing better anonymization strategies for your production databases.")
    input_positional = parser.add_mutually_exclusive_group(required=False)
    input_positional.add_argument("legacy_input",
                        nargs="?",
                        help=argparse.SUPPRESS)

    strategy_positional = parser.add_mutually_exclusive_group(required=False)
    strategy_positional.add_argument("legacy_strategyfile",
                        nargs="?",
                        help=argparse.SUPPRESS)

    output_positional = parser.add_mutually_exclusive_group(required=False)
    output_positional.add_argument("legacy_output",
                        nargs="?",
                        help=argparse.SUPPRESS)

    input_positional.add_argument("--input", "-i",
                        default=os.getenv("PYNONYMIZER_INPUT"),
                        help="The source dump filepath to read from. Use `-` for stdin. [$PYNONYMIZER_INPUT]")

    strategy_positional.add_argument("--strategy", "-s",
                        dest="strategyfile",
                        default=os.getenv("PYNONYMIZER_STRATEGY"),
                        help="A strategyfile to use during anonymization. [$PYNONYMIZER_STRATEGY]")

    output_positional.add_argument("--output", "-o",
                        default=os.getenv("PYNONYMIZER_OUTPUT"),
                        help="The destination filepath to write the dumped output to. Use `-` for stdout. [$PYNONYMIZER_OUTPUT]")

    parser.add_argument("--db-type", "-t",
                        default=os.getenv("PYNONYMIZER_DB_TYPE") or os.getenv("DB_TYPE"),
                        help="Type of database to interact with. More databases will be supported in future versions. default: mysql [$PYNONYMIZER_DB_TYPE]")

    parser.add_argument("--db-host", "-d",
                        default=os.getenv("PYNONYMIZER_DB_HOST") or os.getenv("DB_HOST"),
                        help="Database hostname or IP address. [$PYNONYMIZER_DB_HOST]")

    parser.add_argument("--db-port", "-P",
                        default=os.getenv("PYNONYMIZER_DB_PORT"),
                        help="Database port. Defaults to provider default. [$PYNONYMIZER_DB_PORT]")

    parser.add_argument("--db-name", "-n",
                        default=os.getenv("PYNONYMIZER_DB_NAME") or os.getenv("DB_NAME"),
                        help="Name of database to restore and anonymize in. If not provided, a unique name will be generated from the strategy name. This will be dropped at the end of the run. [$PYNONYMIZER_DB_NAME]")

    parser.add_argument("--db-user", "-u",
                        default=os.getenv("PYNONYMIZER_DB_USER") or os.getenv("DB_USER"),
                        help="Database credentials: username. [$PYNONYMIZER_DB_USER]")

    parser.add_argument("--db-password", "-p",
                        default=os.getenv("PYNONYMIZER_DB_PASSWORD") or os.getenv("DB_PASS"),
                        help="Database credentials: password. Recommended: use environment variables to avoid exposing secrets in production environments. [$PYNONYMIZER_DB_PASSWORD]")

    parser.add_argument("--fake-locale", "-l",
                        default=os.getenv("PYNONYMIZER_FAKE_LOCALE") or os.getenv("FAKE_LOCALE"),
                        help="Locale setting to initialize fake data generation. Affects Names, addresses, formats, etc. [$PYNONYMIZER_FAKE_LOCALE]")

    parser.add_argument("--start-at",
                        default=os.getenv("PYNONYMIZER_START_AT"), dest="start_at_step", choices=ProcessSteps.names(), metavar="STEP",
                        help="Choose a step to begin the process (inclusive). [$PYNONYMIZER_START_AT]")

    parser.add_argument("--only-step",
                        required=False, choices=ProcessSteps.names(), metavar="STEP",
                        default=os.getenv("PYNONYMIZER_ONLY_STEP"), dest="only_step",
                        help="Choose one step to perform. [$PYNONYMIZER_ONLY_STEP]")
                    
    parser.add_argument("--skip-steps",
                        nargs="+", required=False, choices=ProcessSteps.names(), metavar="STEP",
                        default=(lambda: os.getenv("PYNONYMIZER_SKIP_STEPS").split() if os.getenv("PYNONYMIZER_SKIP_STEPS") else [])(), dest="skip_steps",
                        help="Choose one or more steps to skip. [$PYNONYMIZER_SKIP_STEPS]")

    parser.add_argument("--stop-at",
                        default=os.getenv("PYNONYMIZER_STOP_AT"), dest="stop_at_step", choices=ProcessSteps.names(), metavar="STEP",
                        help="Choose a step to stop at (inclusive). [$PYNONYMIZER_STOP_AT]")

    parser.add_argument("--seed-rows",
                        default=os.getenv("PYNONYMIZER_SEED_ROWS"), type=int,
                        help="Specify a number of rows to populate the fake data table used during anonymization.  [$PYNONYMIZER_SEED_ROWS]")

    parser.add_argument("--mssql-driver",
                    default=(os.getenv("PYNONYMIZER_MSSQL_DRIVER")),
                    help="[MSSQL] ODBC driver to use for database connection [$PYNONYMIZER_MSSQL_DRIVER]")


    parser.add_argument("--mssql-backup-compression",
                        action="store_true",
                        default=bool(os.getenv("PYNONYMIZER_MSSQL_BACKUP_COMPRESSION")),
                        help="[MSSQL] Use compression when backing up the database.  [$PYNONYMIZER_MSSQL_BACKUP_COMPRESSION]")

    parser.add_argument("--mysql-cmd-opts",
                    default=os.getenv("PYNONYMIZER_MYSQL_CMD_OPTS"),
                    help="[MYSQL] pass additional arguments to the restore process (advanced use only!).  [$PYNONYMIZER_MYSQL_CMD_OPTS]")
    parser.add_argument("--mysql-dump-opts",
                    default=os.getenv("PYNONYMIZER_MYSQL_DUMP_OPTS"),
                    help="[MYSQL] pass additional arguments to the dump process (advanced use only!).  [$PYNONYMIZER_MYSQL_DUMP_OPTS]")

    parser.add_argument("--postgres-cmd-opts",
                    default=os.getenv("PYNONYMIZER_POSTGRES_CMD_OPTS"),
                    help="[POSTGRES] pass additional arguments to the restore process (advanced use only!).  [$PYNONYMIZER_POSTGRES_CMD_OPTS]")
    parser.add_argument("--postgres-dump-opts",
                    default=os.getenv("PYNONYMIZER_POSTGRES_DUMP_OPTS"),
                    help="[POSTGRES] pass additional arguments to the dump process (advanced use only!).  [$PYNONYMIZER_POSTGRES_DUMP_OPTS]")

    parser.add_argument("-v", "--version", action="version", version=__version__)

    parser.add_argument("--verbose", action="store_true", default=os.getenv("PYNONYMIZER_VERBOSE") or False,
                        help="Increases the verbosity of the logging feature, to help when troubleshooting issues. [$PYNONYMIZER_VERBOSE]"
                        )

    parser.add_argument("--dry-run",
                        default=os.getenv("PYNONYMIZER_DRY_RUN") or False, action="store_true",
                        help="Instruct pynonymizer to skip all process steps. Useful for testing safely.  [$PYNONYMIZER_DRY_RUN]")

    return parser

def _warn_deprecated_env(old_env, new_env):
    logger = logging.getLogger()
    if os.getenv(old_env):
        logger.warning("Environmental var $%s is deprecated. Use $%s", old_env, new_env)

def cli(rawArgs=None):
    """
    Main entry point for the command line. Parse the cmdargs, load env and call to the main process
    :param args:
    :return:
    """
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    logger.handlers.clear()
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(logging.Formatter('%(message)s'))
    logger.addHandler(console_handler)

    # find the dotenv from the current working dir rather than the execution location
    dotenv_file = dotenv.find_dotenv(usecwd=True)
    dotenv.load_dotenv(dotenv_path=dotenv_file)

    parser = create_parser()
    args = parser.parse_args(rawArgs)

    # legacy positionals take precedence if specified
    # This is to support those not using the new options/env fallbacks
    input        = args.legacy_input or args.input
    strategyfile = args.legacy_strategyfile or args.strategyfile
    output = args.legacy_output or args.output

    if args.legacy_input:
        logger.warning("Positional INPUT is deprecated. Use the -i/--input option instead.")
    if args.legacy_strategyfile:
        logger.warning("Positional STRATEGYFILE is deprecated. Use the -s/--strategy option instead.")
    if args.legacy_output:
        logger.warning("Positional OUTPUT is deprecated. Use the -o/--output option instead.")
    if args.fake_locale:
        logger.warning("The locale option -l/--fake-locale is deprecated. Use the locale: key in your strategyfile instead.")

    _warn_deprecated_env("DB_TYPE", "PYNONYMIZER_DB_TYPE")
    _warn_deprecated_env("DB_HOST", "PYNONYMIZER_DB_HOST")
    _warn_deprecated_env("DB_NAME", "PYNONYMIZER_DB_NAME")
    _warn_deprecated_env("DB_USER", "PYNONYMIZER_DB_USER")
    _warn_deprecated_env("DB_PASS", "PYNONYMIZER_DB_PASSWORD")
    _warn_deprecated_env("FAKE_LOCALE", "PYNONYMIZER_FAKE_LOCALE")
    
    if args.verbose:
        console_handler.setLevel(logging.DEBUG)

    # Add local project dir to path in case of custom provider imports     
    if "." not in sys.path:
        sys.path.append(".")
    try:
        pynonymize(
            input_path=input,
            strategyfile_path=strategyfile,
            output_path=output,
            db_type=args.db_type,
            db_host=args.db_host,
            db_port=args.db_port,
            db_name=args.db_name,
            db_user=args.db_user,
            db_password=args.db_password,
            fake_locale=args.fake_locale,
            start_at_step=args.start_at_step,
            only_step=args.only_step,
            skip_steps=args.skip_steps,
            stop_at_step=args.stop_at_step,
            seed_rows=args.seed_rows,
            mssql_driver=args.mssql_driver,
            mssql_backup_compression=args.mssql_backup_compression,
            mysql_cmd_opts=args.mysql_cmd_opts,
            mysql_dump_opts=args.mysql_dump_opts,
            postgres_cmd_opts=args.postgres_cmd_opts,
            postgres_dump_opts=args.postgres_dump_opts,
            dry_run=args.dry_run,
            verbose=args.verbose
        )
    except ModuleNotFoundError as error:
        if error.name == "pyodbc" and args.db_type == "mssql":
            logger.error("Missing Required Packages for database support.")
            logger.error("Install package extras: pip install pynonymizer[mssql]")
            sys.exit(1)
        else:
            raise error
    except ImportError as error:
        if error.name == "pyodbc" and args.db_type == "mssql":
            logger.error("Error importing pyodbc (mssql). "
                         "The ODBC driver may not be installed on your system. See package `unixodbc`.")
            sys.exit(1)
        else:
            raise error
    except DatabaseConnectionError as error:
        logger.error("Failed to connect to database.")
        if args.verbose:
            logger.error(error)
        sys.exit(1)
    except ArgumentValidationError as error:
        logger.error("Missing values for required arguments: \n" + "\n".join(error.validation_messages) +
                     "\nSet these using the command-line options or with environment variables. \n"
                     "For a complete list, See the program help below.\n")
        parser.print_help()
        sys.exit(2)


if __name__ == "__main__":
    cli()



from pynonymizer.database.provider import SEED_TABLE_NAME
from tqdm import tqdm
import logging
from pynonymizer.database.provider import DatabaseProvider
from pynonymizer.database.exceptions import UnsupportedTableStrategyError
from pynonymizer.database.postgres import execution, query_factory
from pynonymizer.database.basic.input import resolve_input
from pynonymizer.database.basic.output import resolve_output
from pynonymizer.strategy.table import TableStrategyTypes


class PostgreSqlProvider(DatabaseProvider):
    """
    A command-line based postgres provider. Uses `psql` and `pg_dump`,
    because of the efficiency of piping mass amounts of sql into the command-line client.
    Unfortunately, this implementation provides limited feedback when things go wrong.
    """
    __CHUNK_SIZE = 8192
    logger = logging.getLogger(__name__)

    def __init__(self, db_host, db_user, db_pass, db_name, db_port=None, seed_rows=None, cmd_opts=None, dump_opts=None):
        if db_port is None:
            db_port = "5432"
        if db_host is None:
            db_host = "127.0.0.1"
        
        if cmd_opts is None:
            cmd_opts = ""
        if dump_opts is None:
            dump_opts = ""

        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port

        if seed_rows is None:
            seed_rows = 150

        self.seed_rows = int(seed_rows)

        self.__runner = execution.PSqlCmdRunner(db_host=db_host, db_user=db_user, db_pass=db_pass, db_name=db_name, db_port=db_port, additional_opts=cmd_opts)
        self.__dumper = execution.PSqlDumpRunner(db_host=db_host, db_user=db_user, db_pass=db_pass, db_name=db_name, db_port=db_port, additional_opts=dump_opts)

    def __seed(self, qualifier_map):
        """
        'Seed' the database with a bunch of pre-generated random records so updates can be performed in batch updates
        """
        for i in tqdm(range(0, self.seed_rows), desc="Inserting seed data", unit="rows"):
            self.logger.debug(f"Inserting seed row {i}")
            self.__runner.db_execute(query_factory.get_insert_seed_row(SEED_TABLE_NAME, qualifier_map))

    def __estimate_dumpsize(self):
        """
        Makes a guess on the dump size using internal database metrics
        :return: A value in bytes, or None (unknown)
        """
        statement = query_factory.get_dumpsize_estimate(self.db_name)
        process_output = self.__runner.get_single_result(statement)

        try:
            return int(process_output)
        except ValueError:
            # Value unparsable, likely NULL
            return None

    def __read_until_empty_byte(self, data):
        return iter(lambda: data.read(self.__CHUNK_SIZE), b'')

    def __run_scripts(self, script_list, title=""):
        for i, script in enumerate(script_list):
            self.logger.info(f"Running {title} script #{i} \"{script[:50]}\"")
            self.logger.info(self.__runner.db_execute(script))

    def create_database(self):
        """Create the working database"""
        self.__runner.execute(query_factory.get_create_database(self.db_name))

    def drop_database(self):
        """Drop the working database"""
        self.__runner.execute(query_factory.get_drop_database(self.db_name))

    def anonymize_database(self, database_strategy):
        """
        Anonymize a restored database using the passed database strategy
        :param database_strategy: a strategy.DatabaseStrategy configuration
        :return:
        """
        qualifier_map = database_strategy.fake_update_qualifier_map

        if len(qualifier_map) > 0:
            self.logger.info("creating seed table with %d columns", len(qualifier_map))
            create_seed_table_sql = query_factory.get_create_seed_table(SEED_TABLE_NAME, qualifier_map)
            self.__runner.db_execute(create_seed_table_sql)

            self.logger.info("Inserting seed data")
            self.__seed(qualifier_map)

        self.__run_scripts(database_strategy.before_scripts, "before")

        table_strategies = database_strategy.table_strategies
        self.logger.info("Anonymizing %d tables", len(table_strategies))

        with tqdm(desc="Anonymizing database", total=len(table_strategies)) as progressbar:
            for table_strategy in table_strategies:
                if table_strategy.strategy_type == TableStrategyTypes.TRUNCATE:
                    progressbar.set_description("Truncating {}".format(table_strategy.qualified_name))
                    self.__runner.db_execute(query_factory.get_truncate_table(table_strategy))

                elif table_strategy.strategy_type == TableStrategyTypes.DELETE:
                    progressbar.set_description("Deleting {}".format(table_strategy.qualified_name))
                    self.__runner.db_execute(query_factory.get_delete_table(table_strategy))

                elif table_strategy.strategy_type == TableStrategyTypes.UPDATE_COLUMNS:
                    progressbar.set_description("Anonymizing {}".format(table_strategy.qualified_name))
                    statements = query_factory.get_update_table(SEED_TABLE_NAME, table_strategy)
                    self.__runner.db_execute(statements)

                else:
                    raise UnsupportedTableStrategyError(table_strategy)

                progressbar.update()

        self.__run_scripts(database_strategy.after_scripts, "after")

        self.logger.info("dropping seed table")
        self.__runner.db_execute(query_factory.get_drop_seed_table(SEED_TABLE_NAME))

    def restore_database(self, input_path):
        """
        Feed a `pg_dump` dumpfile to the `psql` binary on stdin.
        :param input_path:
        :return:
        """
        input_obj = resolve_input(input_path)
        dumpsize = input_obj.get_size()

        batch_processor = self.__runner.open_batch_processor()
        with input_obj.open() as dumpfile_data:
            with tqdm(desc="Restoring", total=dumpsize, unit='B', unit_scale=True, unit_divisor=1000) as bar:
                for chunk in self.__read_until_empty_byte(dumpfile_data):
                    batch_processor.write(chunk)
                    batch_processor.flush()
                    bar.update(len(chunk))

    def dump_database(self, output_path):
        """
        Feed an output with stdout from the dump binary
        :param output_path:
        :return:
        """
        output_obj = resolve_output(output_path)
        dumpsize_estimate = self.__estimate_dumpsize()

        dump_process = self.__dumper.open_dumper()
        with output_obj.open() as output_file:
            with tqdm(desc="Dumping", total=dumpsize_estimate, unit='B', unit_scale=True, unit_divisor=1000) as bar:
                for chunk in self.__read_until_empty_byte(dump_process):
                    output_file.write(chunk)
                    bar.update(len(chunk))



import shutil
import shlex
import subprocess
from pynonymizer.database.exceptions import DependencyError
"""
Seperate everything that touches actual query exec into its own module
"""

class MySqlDumpRunner:
    def __init__(self, db_host, db_user, db_pass, db_name, db_port="3306", additional_opts = ""):
        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port
        self.additional_opts = shlex.split(additional_opts)

        if not (shutil.which("mysqldump")):
            raise DependencyError( "mysqldump", "The 'mysqldump' client must be present in the $PATH")

    def __get_base_params(self):
        return ["--host", self.db_host, "--port", self.db_port, "--user", self.db_user, f"-p{self.db_pass}"]

    def open_dumper(self):
        return subprocess.Popen(["mysqldump"] + self.__get_base_params() + self.additional_opts + [self.db_name], stdout=subprocess.PIPE).stdout


class MySqlCmdRunner:
    def __init__(self, db_host, db_user, db_pass, db_name, db_port="3306", additional_opts = ""):
        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port
        self.additional_opts = shlex.split(additional_opts)

        if not (shutil.which("mysql")):
            raise DependencyError("mysql", "The 'mysql' client must be present in the $PATH")


    def __mask_subprocess_error(self, error):
        """
        messes with the internals of a CalledProcessError to hide the fact that there's a password in there,
        in case it bubbles out in a traceback.

        This might be better as a wrapping exception, rather than messing around inside other people's classes.
        """
        error.cmd = ["mysql", "-h", self.db_host, "-P", self.db_port, "-u", self.db_user, "-p******"]
        raise error from None

    def __get_base_params(self):
        return ["mysql", "-h", self.db_host, "-P", self.db_port, "-u", self.db_user, f"-p{self.db_pass}"]

    def execute(self, statements):
        if not isinstance(statements, list):
            statements = [statements]

        outputs = []

        for statement in statements:
            try:
                outputs.append( subprocess.check_output(self.__get_base_params() + self.additional_opts + ["--execute", statement]) )
            except subprocess.CalledProcessError as error:
                self.__mask_subprocess_error(error)

        return outputs

    def db_execute(self, statements):
        if not isinstance(statements, list):
            statements = [statements]

        outputs = [];

        for statement in statements:
            try:
                outputs.append( subprocess.check_output(self.__get_base_params() + self.additional_opts + [self.db_name,  "--execute", statement]) )
            except subprocess.CalledProcessError as error:
                self.__mask_subprocess_error(error)

        return outputs

    def get_single_result(self, statement):
        try:
            return subprocess.check_output(self.__get_base_params() + ["-sN", *self.additional_opts, self.db_name, "--execute", statement]).decode()
        except subprocess.CalledProcessError as error:
            self.__mask_subprocess_error(error)

    def open_batch_processor(self):
        return subprocess.Popen(self.__get_base_params() + self.additional_opts + [self.db_name], stdin=subprocess.PIPE).stdin




from pynonymizer.fake import FakeColumnGenerator
from pynonymizer.strategy.exceptions import UnknownColumnStrategyError, UnknownTableStrategyError, ConfigSyntaxError
from pynonymizer.strategy.table import UpdateColumnsTableStrategy, TruncateTableStrategy, DeleteTableStrategy, TableStrategyTypes
from pynonymizer.strategy.update_column import (
    UpdateColumnStrategyTypes,
    EmptyUpdateColumnStrategy,
    UniqueEmailUpdateColumnStrategy,
    UniqueLoginUpdateColumnStrategy,
    FakeUpdateColumnStrategy,
    LiteralUpdateColumnStrategy
)
from pynonymizer.strategy.database import DatabaseStrategy
from copy import deepcopy
import logging

logger = logging.getLogger(__name__)


class StrategyParser:
    def __init__(self, fake_locale):
        self.fake_locale = fake_locale

    @staticmethod
    def __normalize_table_config(table_config):
        # If the table config doesn't have a "type" specified, it needs to be determined automatically.
        if "type" not in table_config:
            if "columns" in table_config:
                return {
                    **table_config,
                    "type": TableStrategyTypes.UPDATE_COLUMNS.value,
                }

            elif table_config == "truncate":
                return {
                    "type": TableStrategyTypes.TRUNCATE.value,
                }
            
            elif table_config == "delete":
                return {
                    "type": TableStrategyTypes.DELETE.value,
                }

            else:
                raise UnknownTableStrategyError(table_config)

        else:
            return table_config

    @staticmethod
    def __normalize_column_config(column_config):
        # If the column config doesn't have a "type" specified, it needs to be determined automatically.
        if "type" not in column_config:
            if column_config == "empty":
                return {
                    "type": UpdateColumnStrategyTypes.EMPTY.value
                }

            elif column_config == "unique_email":
                return {
                    "type": UpdateColumnStrategyTypes.UNIQUE_EMAIL.value
                }

            elif column_config == "unique_login":
                return {
                    "type": UpdateColumnStrategyTypes.UNIQUE_LOGIN.value
                }
            elif column_config.startswith("(") and column_config.endswith(")"):
                return {
                    "type": UpdateColumnStrategyTypes.LITERAL.value,
                    "value": column_config
                }

            else:
                # Can't match it to anything special, must be a fake_update type. move the value to the fake_type field
                return {
                    "type": UpdateColumnStrategyTypes.FAKE_UPDATE.value,
                    "fake_type": column_config
                }
        else:
            return column_config

    @staticmethod
    def __normalize_table_list(config):
        # if tables is given in dict form, normalize to list with table_name key
        if "tables" in config and isinstance(config["tables"], dict):
            tables_list = []
            for table_name, table_config in config["tables"].items():
                normalized_table = StrategyParser.__normalize_table_config(table_config)
                normalized_table["table_name"] = table_name
                tables_list.append(normalized_table)

            config["tables"] = tables_list

        return config

    @staticmethod
    def __normalize_update_columns_list(columns_config):
        # if columns is given in dict form, normalize to list
        if isinstance(columns_config, dict):
            column_list = []
            for column_name, column_config in columns_config.items():
                normalized_column = StrategyParser.__normalize_column_config(column_config)
                normalized_column["column_name"] = column_name
                column_list.append( normalized_column )

            return column_list

        elif isinstance(columns_config, list):
            return [StrategyParser.__normalize_column_config(col) for col in columns_config]
        else:
            raise ConfigSyntaxError("Unknown update column config syntax: {}".format(columns_config))


    def __parse_update_column(self, column_config):
        update_column_type = UpdateColumnStrategyTypes.from_value(column_config.pop("type"))
        try:
            if update_column_type == UpdateColumnStrategyTypes.EMPTY:
                return EmptyUpdateColumnStrategy(**column_config)

            elif update_column_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN:
                return UniqueLoginUpdateColumnStrategy(**column_config)

            elif update_column_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL:
                return UniqueEmailUpdateColumnStrategy(**column_config)

            elif update_column_type == UpdateColumnStrategyTypes.FAKE_UPDATE:
                return FakeUpdateColumnStrategy(fake_column_generator=self.fake_seeder, **column_config)

            elif update_column_type == UpdateColumnStrategyTypes.LITERAL:
                return LiteralUpdateColumnStrategy(**column_config)

            else:
                raise UnknownColumnStrategyError(column_config)
        except TypeError as error:
            # TypeError can be thrown when the dict args dont match the constructors for the types. We need to re-throw
            raise ConfigSyntaxError()

    def __parse_table(self, table_config):
        table_strategy = TableStrategyTypes.from_value(table_config.pop("type"))

        try:
            if table_strategy == TableStrategyTypes.TRUNCATE:
                return TruncateTableStrategy(**table_config)
            elif table_strategy == TableStrategyTypes.DELETE:
                return DeleteTableStrategy(**table_config)
            elif table_strategy == TableStrategyTypes.UPDATE_COLUMNS:
                # update columns supports dict and list columns, so this has to be normalized again during parsing.
                normalized_columns = StrategyParser.__normalize_update_columns_list(table_config.pop("columns"))
                parsed_columns = [self.__parse_update_column(column) for column in normalized_columns]

                return UpdateColumnsTableStrategy(column_strategies=parsed_columns, **table_config)
            else:
                raise UnknownTableStrategyError(table_config)

        except TypeError as error:
            # TypeError can be thrown when the dict args dont match the constructors for the types. We need to re-throw
            raise ConfigSyntaxError()



    def parse_config(self, raw_config, locale_override=None):
        """
        parse a configuration dict into a DatabaseStrategy.
        :param raw_config:
        :return:
        """
        config = StrategyParser.__normalize_table_list(deepcopy(raw_config))

        locale = config.get("locale", None)
        if locale_override:
            locale = locale_override

        providers = config.get("providers", [])
        self.fake_seeder = FakeColumnGenerator(locale=locale, providers=providers)

        table_strategies = [self.__parse_table(table_config) for table_config in config["tables"]]

        before_scripts = None
        after_scripts = None
        try:
            scripts = config["scripts"]

            if "before" in scripts:
                before_scripts = scripts["before"]

            if "after" in scripts:
                after_scripts = scripts["after"]
        except KeyError:
            pass

        return DatabaseStrategy(
            table_strategies=table_strategies,
            before_scripts=before_scripts,
            after_scripts=after_scripts
        )



from abc import ABC, abstractmethod

SEED_TABLE_NAME = "_pynonymizer_seed_fake_data"

class DatabaseProvider(ABC):
    @abstractmethod
    def create_database(self):
        pass

    @abstractmethod
    def drop_database(self):
        pass

    @abstractmethod
    def anonymize_database(self, database_strategy):
        pass

    @abstractmethod
    def restore_database(self, input_obj):
        pass

    @abstractmethod
    def dump_database(self, output_obj):
        pass



"""
database.basic is not a real provider but it contains common components for providers that share a common core
and rough strategy:

e.g. mysql and postgres share a ansi sql subprocess-based mysql/psql/mysqldump/pgdump interface

"""


import os
import uuid
from pynonymizer.database.mysql import MySqlProvider
from pynonymizer.database.mssql import MsSqlProvider
from pynonymizer.database.postgres import PostgreSqlProvider
from pynonymizer.database.exceptions import UnknownDatabaseTypeError

def get_temp_db_name(filename=None):
    name, _ = os.path.splitext(os.path.basename(filename))
    return f"{name}_{uuid.uuid4().hex}"


def get_provider(type, *args, **kwargs):
    provider = None

    if type == "mysql":
        provider = MySqlProvider
    if type == "mssql":
        provider = MsSqlProvider
    elif type == "postgres":
        provider = PostgreSqlProvider

    if provider:
        return provider(*args, **kwargs)
    else:
        raise UnknownDatabaseTypeError(type)



from pynonymizer.exceptions import PynonymizerError


class StrategyError(PynonymizerError):
    pass


class UnknownUpdateColumnFakeTypeError(StrategyError):
    def __init__(self, config):
        self.config = config
        super().__init__("Unknown fake column type: {}".format(config))


class UnknownTableStrategyError(StrategyError):
    def __init__(self, config):
        self.config = config
        super().__init__("Unable to determine table strategy from value: {}".format(config))


class UnknownColumnStrategyError(StrategyError):
    def __init__(self, config):
        self.config = config
        super().__init__("Unable to determine column strategy from value: {}".format(config))


class ConfigSyntaxError(StrategyError):
    pass



import gzip
import lzma
import os
import sys
"""
Streaming output files for streamable providers
"""


class UnknownOutputTypeError(Exception):
    def __init__(self, filename):
        super().__init__("Unable to detect output type for file: {}".format(filename))


class GzipOutput:
    def __init__(self, filename):
        self.filename = filename

    def open(self):
        return gzip.open(self.filename, "wb")

class XzOutput:
    def __init__(self, filename):
        self.filename = filename

    def open(self):
        return lzma.open(self.filename, "wb")

class RawOutput:
    def __init__(self, filename):
        self.filename = filename

    def open(self):
        return open(self.filename, "wb")

class StdOutOutput:
    def open(self):
        return sys.stdout.buffer

def resolve_output(filename):
    if filename == "-":
        return StdOutOutput()

    name, ext = os.path.splitext(filename)

    if ext == ".sql":
        return RawOutput(filename)
    elif ext == ".gz":
        return GzipOutput(filename)
    elif ext == ".xz":
        return XzOutput(filename)
    else:
        raise UnknownOutputTypeError(filename)


import logging
from pynonymizer.database import get_temp_db_name, get_provider
from pynonymizer.fake import FakeColumnGenerator
from pynonymizer.strategy.parser import StrategyParser
from pynonymizer.strategy.config import read_config
from pynonymizer.exceptions import ArgumentValidationError, DatabaseConnectionError
from pynonymizer.process_steps import StepActionMap, ProcessSteps


logger = logging.getLogger(__name__)


def pynonymize(
        input_path=None, strategyfile_path=None, output_path=None, db_user=None, db_password=None, db_type=None,
        db_host=None, db_name=None, db_port=None, fake_locale=None, only_step=None, start_at_step=None, stop_at_step=None, skip_steps=None,
        seed_rows=None, dry_run=False, verbose=False,

        **kwargs
    ):
    """
    Runs a pynonymize process as if the CLI had been invoked.

    :raises:
        ArgumentValidationError: used when kwargs are missing or unable to be auto-resolved.
        
    """
    # Default and Normalize args
    if only_step is not None:
        only_step = ProcessSteps.from_value(only_step)

    if start_at_step is None:
        start_at_step = ProcessSteps.START
    else:
        start_at_step = ProcessSteps.from_value(start_at_step)

    if stop_at_step is None:
        stop_at_step = ProcessSteps.END
    else:
        stop_at_step = ProcessSteps.from_value(stop_at_step)

    if skip_steps and len(skip_steps) > 0:
        skip_steps = [ProcessSteps.from_value(skip) for skip in skip_steps]

    if db_type is None:
        db_type = "mysql"

    if fake_locale is None:
        fake_locale = "en_GB"

    if seed_rows is None:
        seed_rows = 150

    actions = StepActionMap(
        start_at_step=start_at_step,
        stop_at_step=stop_at_step,
        skip_steps=skip_steps,
        dry_run=dry_run,
        only_step=only_step
    )

    # Validate mandatory args (depends on step actions)
    validations = []

    if not actions.skipped(ProcessSteps.RESTORE_DB):
        if input_path is None:
            validations.append("Missing INPUT")

    if not actions.skipped(ProcessSteps.ANONYMIZE_DB):
        if strategyfile_path is None:
            validations.append("Missing STRATEGYFILE")
        else:
            # only auto-determine the db_name if we have a strategyfile AND we are anonymizing.
            if db_name is None:
                db_name = get_temp_db_name(strategyfile_path)

    if not actions.skipped(ProcessSteps.DUMP_DB):
        if output_path is None:
            validations.append("Missing OUTPUT")

    if db_user is None:
        validations.append("Missing DB_USER")

    if db_password is None:
        validations.append("Missing DB_PASSWORD")

    if db_name is None:
        validations.append("Missing DB_NAME: Auto-resolve failed.")

    if len(validations) > 0:
        raise ArgumentValidationError(validations)

    # init strategy as it relies on I/O - fail fast here preferred to after restore
    if not actions.skipped(ProcessSteps.ANONYMIZE_DB):
        strategy_parser = StrategyParser(fake_locale)

        logger.debug("loading strategyfile %s...", strategyfile_path)
        strategy = strategy_parser.parse_config(read_config(strategyfile_path), locale_override=fake_locale)

    # Discover db-type kwargs
    # mssql_backup_option -> backup_option and pass these to the constructor
    db_kwargs = {}
    db_arg_prefix = f"{db_type}_"
    for k, v in kwargs.items():
        if k.startswith(db_arg_prefix):
            db_kwargs[ k[len(db_arg_prefix):] ] = v

    logger.debug("Database: (%s:%s)%s@%s name: %s", db_host, db_port, db_type, db_user, db_name)
    db_provider = get_provider(
        type=db_type,
        db_host=db_host,
        db_user=db_user,
        db_pass=db_password,
        db_name=db_name,
        db_port=db_port,
        seed_rows=seed_rows,
        **db_kwargs
    )

    # main process - no destructive/non-retryable actions should happen before this line ---
    logger.info(actions.summary(ProcessSteps.CREATE_DB))
    if not actions.skipped(ProcessSteps.CREATE_DB):
        db_provider.create_database()

    logger.info(actions.summary(ProcessSteps.RESTORE_DB))
    if not actions.skipped(ProcessSteps.RESTORE_DB):
        db_provider.restore_database(input_path)

    logger.info(actions.summary(ProcessSteps.ANONYMIZE_DB))
    if not actions.skipped(ProcessSteps.ANONYMIZE_DB):
        db_provider.anonymize_database(strategy)

    logger.info(actions.summary(ProcessSteps.DUMP_DB))
    if not actions.skipped(ProcessSteps.DUMP_DB):
        db_provider.dump_database(output_path)

    logger.info(actions.summary(ProcessSteps.DROP_DB))
    if not actions.skipped(ProcessSteps.DROP_DB):
        db_provider.drop_database()

    logger.info("Process complete!")



import shutil
import shlex
import subprocess
from pynonymizer.database.exceptions import DependencyError
import os
"""
Seperate everything that touches actual query exec into its own module
"""

class PSqlDumpRunner:
    def __init__(self, db_host, db_user, db_pass, db_name, db_port="5432", additional_opts = ""):
        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port
        self.additional_opts = shlex.split(additional_opts)

        if not (shutil.which("pg_dump")):
            raise DependencyError( "pg_dump", "The 'pg_dump' client must be present in the $PATH")

    def __get_base_params(self):
        return ["pg_dump", "--host", self.db_host, "--port", self.db_port, "--username", self.db_user]

    def __get_env(self):
        new_env = os.environ.copy()
        new_env.update({"PGPASSWORD": self.db_pass})

        return new_env

    def open_dumper(self):
        return subprocess.Popen(self.__get_base_params() + self.additional_opts + [self.db_name],
                                stdout=subprocess.PIPE,
                                env=self.__get_env()
                                ).stdout


class PSqlCmdRunner:
    def __init__(self, db_host, db_user, db_pass, db_name, db_port="5432", additional_opts = ""):
        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port
        self.additional_opts = shlex.split(additional_opts)

        if not (shutil.which("psql")):
            raise DependencyError("psql", "The 'psql' client must be present in the $PATH")

    def __get_base_params(self):
        return ["psql", "--host", self.db_host, "--port", self.db_port, "--username", self.db_user]

    def __get_env(self):
        new_env = os.environ.copy()
        new_env.update({"PGPASSWORD": self.db_pass})

        return new_env

    def execute(self, statements):
        if not isinstance(statements, list):
            statements = [statements]

        outputs = []

        for statement in statements:
            outputs.append( subprocess.check_output(self.__get_base_params() + self.additional_opts + ["--command", statement], env=self.__get_env()) )

        return outputs

    def db_execute(self, statements):
        if not isinstance(statements, list):
            statements = [statements]

        outputs = []
        
        for statement in statements:
            outputs.append(subprocess.check_output(
                self.__get_base_params() + ["--dbname", self.db_name, *self.additional_opts,  "--command", statement],
                env=self.__get_env()
            ))

        return outputs

    def get_single_result(self, statement):
        return subprocess.check_output(
            self.__get_base_params() + ["--dbname", self.db_name, "-tA", *self.additional_opts, "--command", statement],
            env=self.__get_env()
        ).decode()


    def open_batch_processor(self):
        return subprocess.Popen(
            self.__get_base_params() + ["--dbname", self.db_name, "--quiet"] + self.additional_opts,
            env=self.__get_env(),
            stdin=subprocess.PIPE
        ).stdin



import os
import struct
import gzip
import sys
"""
Streaming input files for streamable providers
"""


class UnknownInputTypeError(Exception):
    def __init__(self, filename):
        super().__init__("Unable to detect input type for file: {}".format(filename))


class GzipInput:
    def __init__(self, filename):
        self.filename = filename

    def get_size(self):
        """
        Get the uncompressed filesize of a gzip
        The last 4 bytes of a gzip contain the uncompressed filesize in little endian format (broken on >4GB zips)
        :param filename: gzip filename to read from
        :returns uncompressed gzip filesize.
        """
        with open(self.filename, 'rb') as file:
            file.seek(-4, 2)
            size = file.read()
            return struct.unpack('<I', size)[0]

    def open(self):
        return gzip.open(self.filename, "rb")


class RawInput:
    def __init__(self, filename):
        self.filename = filename

    def get_size(self):
        return os.path.getsize(self.filename)

    def open(self):
        return open(self.filename, "rb")

class StdInInput:
    def get_size(self):
        return None

    def open(self):
        return sys.stdin.buffer


def resolve_input(filename):
    if filename == "-":
        return StdInInput()

    name, ext = os.path.splitext(filename)

    if ext == ".sql":
        return RawInput(filename)
    elif ext == ".gz":
        return GzipInput(filename)
    else:
        raise UnknownInputTypeError(filename)


import setuptools
import sys

if sys.version_info < (3, 6):
    sys.exit('pynonymizer requires Python 3.6+ to run')

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='pynonymizer',
    description='An anonymization tool for production databases',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Topic :: Database',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    author='Jerome Twell',
    author_email='jtwell1@gmail.com',
    url="https://github.com/jerometwell/pynonymizer",
    keywords="anonymization gdpr database mysql",
    license="MIT",
    packages=setuptools.find_packages(exclude=("tests", "tests.*")),
    install_requires=[
        "pyyaml>=5",
        "tqdm>=4",
        "faker>=1",
        "python-dotenv>=0.10",
    ],
    entry_points={
        'console_scripts': [
            'pynonymizer = pynonymizer.cli:cli'
        ]
    },
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    extras_require={
        'mssql': ["pyodbc>=4.0.26"]
    }
 )



from enum import Enum


class ProcessSteps(Enum):
    START = 0
    GET_SOURCE = 100
    CREATE_DB = 200
    RESTORE_DB = 300
    ANONYMIZE_DB = 400
    DUMP_DB = 500
    DROP_DB = 600
    END = 9999

    @staticmethod
    def names():
        return [step.name for step in ProcessSteps]

    @staticmethod
    def from_value(step_value):
        """
        resolve a enum value from key (case insensitive)
        :return: ProcessSteps property
        """
        # Try to resolve as a string value
        return ProcessSteps[step_value.upper()]


class SkipReason:
    pass


class StepBeforeStartReason(SkipReason):
    def __init__(self, start_at_step):
        self.start_at_step = start_at_step

    def __str__(self):
        return "Starting at [{}]".format(self.start_at_step.name)


class StepAfterStopReason(SkipReason):
    def __init__(self, stop_at_step):
        self.stop_at_step = stop_at_step

    def __str__(self):
        return "Stopped at [{}]".format(self.stop_at_step.name)


class StepSkippedReason(SkipReason):
    def __init__(self, skipped_steps):
        self.skipped_steps = skipped_steps

    def __str__(self):
        formatted_skips = ", ".join(["[{}]".format(step.name) for step in self.skipped_steps])
        return "Skipping ({})".format(formatted_skips)

class DryRunReason(SkipReason):
    def __str__(self):
        return "Skipping (DRY RUN)"

class StepAction:
    """
    A container that represents the action for a step (run, or skip) and the reason(s) why
    """

    def __init__(self, process_step, start_at_step, stop_at_step, skip_steps, dry_run=False):
        self.process_step = process_step

        skip_reasons = []
        if dry_run:
            skip_reasons.append(DryRunReason())

        if start_at_step.value > process_step.value:
            skip_reasons.append(StepBeforeStartReason(start_at_step))

        if stop_at_step.value < process_step.value:
            skip_reasons.append(StepAfterStopReason(stop_at_step))

        if process_step in skip_steps:
            skip_reasons.append(StepSkippedReason(skip_steps))

        if len(skip_reasons) > 0:
            self.skipped = True
            self.skip_reasons = skip_reasons
        else:
            self.skipped = False

    def __eq__(self, other):
        return self.skipped == other.skipped and self.skip_reasons == other.skip_reasons

    @property
    def summary(self):
        if self.skipped:
            skip_strings = [str(reason) for reason in self.skip_reasons]
            return "Skipped [{}]: ({})".format(self.process_step.name, ",\n".join(skip_strings))
        else:
            return "[{}]".format(self.process_step.name)


class StepActionMap:
    def __init__(self, start_at_step=ProcessSteps.START, stop_at_step=ProcessSteps.END, skip_steps=None, dry_run=False, only_step=None):
        action_map = {}
        if only_step:
            start_at_step = only_step
            stop_at_step = only_step
            
        if skip_steps is None:
            skip_steps = []
        for step in ProcessSteps:
            action_map[step] = StepAction(step, start_at_step, stop_at_step, skip_steps, dry_run=dry_run)

        self.__action_map = action_map

    def step(self, step):
        return self.__action_map[step]

    def skipped(self, step):
        return self.__action_map[step].skipped

    def summary(self, step):
        return self.__action_map[step].summary

    def any_skipped(self, *steps):
        for step in steps:
            if self.skipped(step):
                return True

        return False

    def all_skipped(self, *steps):
        for step in steps:
            if not self.skipped(step):
                return False

        return True



from faker import Faker
import logging
import inspect
from enum import Enum
import importlib

logger = logging.getLogger(__name__)


class FakeDataType(Enum):
    STRING = "STRING"
    DATE = "DATE"
    DATETIME = "DATETIME"
    INT = "INT"

"""
We have to create a mapping here between faker types and rough return values, because providers need to know how to
Store the generated values in the 'seed' table.

FakeColumnGenerator will assume STRING unless the below mapping overrides it.
"""
_FAKE_DATA_TYPES = {
    # date_time
    "random_int":      FakeDataType.INT,
    "date_between":   FakeDataType.DATE,
    "date_between_dates":   FakeDataType.DATE,
    "date_object":   FakeDataType.DATE,
    "date_of_birth":   FakeDataType.DATE,
    "date_this_century":   FakeDataType.DATE,
    "date_this_decade":   FakeDataType.DATE,
    "date_this_month":   FakeDataType.DATE,
    "date_this_year":   FakeDataType.DATE,
    "date_time":   FakeDataType.DATETIME,
    "date_time_ad":   FakeDataType.DATETIME,
    "date_time_between":   FakeDataType.DATETIME,
    "date_time_between_dates":   FakeDataType.DATETIME,
    "date_time_this_century":   FakeDataType.DATETIME,
    "date_time_this_decade":   FakeDataType.DATETIME,
    "date_time_this_month":   FakeDataType.DATETIME,
    "date_time_this_year":   FakeDataType.DATETIME,
    "future_date":   FakeDataType.DATE,
    "future_datetime":   FakeDataType.DATETIME,
    "past_date":   FakeDataType.DATE,
    "past_datetime":   FakeDataType.DATETIME,
    "date":            FakeDataType.DATE
}


class UnsupportedFakeTypeError(Exception):
    def __init__(self, fake_type, kwargs=None):
        kwargs = {} if kwargs is None else kwargs
        super().__init__(f"Unsupported Fake type: {fake_type}: {kwargs.keys()}")
        self.fake_type = fake_type
        self.kwargs = kwargs


class FakeColumnGenerator:
    def __init__(self, locale=None, providers=[]):
        self.__faker = Faker(locale)

        for provider_path in providers:
            module_path, cls_name = provider_path.rsplit(".", 1)
            imported = importlib.import_module(module_path)
            imported_cls = getattr(imported, cls_name)
            self.__faker.add_provider(imported_cls)

    def supports(self, method_name, additional_kwargs=None):
        additional_kwargs = {} if additional_kwargs is None else additional_kwargs
        try:
            method = getattr(self.__faker, method_name)
        except AttributeError:
            return False

        argspec = inspect.getfullargspec(method)
        if all(kwarg in argspec.args for kwarg in additional_kwargs):
            return True
        else:
            return False

    def get_data_type(self, method_name):
        """Get the datatype for a column's method_name"""
        try:
            return _FAKE_DATA_TYPES[method_name]
        except KeyError:
            return FakeDataType.STRING

    def get_value(self, method_name, additional_kwargs=None):
        """
        Get a value from a fake column using the appropriate faker method and it's kwargs
        """
        additional_kwargs = {} if additional_kwargs is None else additional_kwargs

        if not self.supports(method_name, additional_kwargs):
            raise UnsupportedFakeTypeError(method_name, additional_kwargs)

        method = getattr(self.__faker, method_name)

        return method(**additional_kwargs)





from pynonymizer.exceptions import PynonymizerError


class DatabaseProviderError(PynonymizerError):
    pass

class DependencyError(DatabaseProviderError):
    def __init__(self, name, message):
        self.name = name
        super().__init__("Bad dependency: {}, {}".format(name, message))


class UnsupportedTableStrategyError(DatabaseProviderError):
    def __init__(self, table_strategy):
        super().__init__("Unsupported Table Strategy: {}".format(table_strategy))


class UnknownDatabaseTypeError(DatabaseProviderError):
    def __init__(self, database_type):
        self.database_type = database_type
        super().__init__("Unknown Database Type: {}".format(database_type))


class UnsupportedColumnStrategyError(DatabaseProviderError):
    def __init__(self, column_strategy):
        self.column_strategy = column_strategy
        super().__init__("Unsupported Column Strategy Type: {}".format(column_strategy))